#!/usr/bin/env python

import os, subprocess
import re

def src(level):
    return "CompCert2.4"
def tgt(level):
    return "level%s" % level

PASS_CLASSIFIERS = [
    (r"driver/Compiler.v", "driver/Compiler.v", []),
    (r"backend/ValueAnalysis.v", ".../ValueAnalysis.v", []),
    (r"cfrontend/Cstrategy.v", ".../Cstrategy.v", []),
    (r"cfrontend/SimplExpr(\w+).v", ".../SimplExpr(proof|spec).v", [3]),
    (r"cfrontend/SimplLocalsproof.v", ".../SimplLocalsproof.v", []),
    (r"cfrontend/Cshmgenproof.v", ".../Cshmgenproof.v", []),
    (r"cfrontend/Cminorgenproof.v", ".../Cminorgenproof.v", []),
    (r"backend/Select(\w+)proof.v", ".../Select*proof.v", [3]),
    (r"backend/RTLgen(\w+).v", ".../RTLgen(proof|spec).v", []),
    (r"backend/Tailcallproof.v", ".../Tailcallproof.v", [5]),
    (r"backend/Inlining(\w+).v", ".../Inlining(proof|spec).v", [3, 5]),
    (r"backend/Renumberproof.v", ".../Renumberproof.v", [5]),
    (r"backend/Constpropproof.v", ".../Constpropproof.v", [3, 5]),
    (r"backend/CSEproof.v", ".../CSEproof.v", [3, 5]),
    (r"backend/Deadcodeproof.v", ".../Deadcodeproof.v", [3, 5]),
    (r"backend/Allocproof.v", ".../Allocproof.v", []),
    (r"backend/Tunnelingproof.v", ".../Tunnelingproof.v", []),
    (r"backend/Linearizeproof.v", ".../Linearizeproof.v", []),
    (r"backend/CleanupLabelsproof.v", ".../CleanupLabelsproof.v", []),
    (r"backend/Stackingproof.v", ".../Stackingproof.v", []),
    (r"backend/Asmgenproof0.v", ".../Asmgenproof0proof.v", []),
    (r"arm/(\w+)proof(\w*).v", "arm/*proof*.v", []),
    (r"ia32/(\w+)proof(\w*).v", "ia32/*proof*.v", []),
    (r"powerpc/(\w+)proof(\w*).v", "powerpc/*proof*.v", []),
]

METATHEORY_CLASSIFIERS = [
]

# "backend/RTLExtra.v|driver/CompilerExtra.v|driver/Linker.v|driver/Linkerproof.v|"

def check_output(*popenargs, **kwargs):
    if 'stdout' in kwargs:
        raise ValueError('stdout argument not allowed, it will be overridden.')
    process = subprocess.Popen(stdout=subprocess.PIPE, *popenargs, **kwargs)
    output, unused_err = process.communicate()
    retcode = process.poll()
    return (retcode, output)

def add_lines(x, y):
    return (x[0] + y[0], x[1] + y[1], x[2] + y[2], x[3] + y[3])

def acc_lines(l):
    acc = (0, 0, 0, 0)
    for x in l:
        acc = add_lines(acc, x[1])
    return acc

def add_lines7(x, y):
    return (x[0] + y[0], x[1] + y[1], x[2] + y[2], x[3] + y[3], x[4] + y[4], x[5] + y[5], x[6] + y[6])

def acc_lines7(l):
    acc = (0, 0, 0, 0, 0, 0, 0)
    for pat in l:
        acc = add_lines7(acc, l[pat])
    return acc

def nonzerostr(x):
    if x == 0:
        return ""
    else:
        return str(x)

def str_of_ratio(x, y):
    if x == 0 or y == 0:
        return ""
    else:
        return "%2.2lf" % (100.0 * float(y) / float(x))

def str_of_tex_ratio(x, y):
    if x == 0 or y == 0:
        return "\\phantom{ (00.0\\%)}"
    else:
        return " ({:4.1f}\\%)".format(100.0 * float(y) / float(x)).replace("( ", "\\phantom{0}(")

def str_of_lines(l):
    return ("\t".join(map(lambda x: nonzerostr(x), l))) + "\t" + ("\t".join(map(lambda x: str_of_ratio(l[0], x), l[1:])))

def str_of_tex_lines(l, shades):
    return "%s%s & %s%s%s & %s%s%s & %s%s%s & %s%s%s & %s%s%s & %s%s%s" % \
        ("\\cellcolor[gray]{0.8}" if (0 in shades) else "", nonzerostr(l[0]),
         "\\cellcolor[gray]{0.8}" if (1 in shades) else "", nonzerostr(l[1]), str_of_tex_ratio(l[0], l[1]),
         "\\cellcolor[gray]{0.8}" if (2 in shades) else "", nonzerostr(l[2]), str_of_tex_ratio(l[0], l[2]),
         "\\cellcolor[gray]{0.8}" if (3 in shades) else "", nonzerostr(l[3]), str_of_tex_ratio(l[0], l[3]),
         "\\cellcolor[gray]{0.8}" if (4 in shades) else "", nonzerostr(l[4]), str_of_tex_ratio(l[0], l[4]),
         "\\cellcolor[gray]{0.8}" if (5 in shades) else "", nonzerostr(l[5]), str_of_tex_ratio(l[0], l[5]),
         "\\cellcolor[gray]{0.8}" if (6 in shades) else "", nonzerostr(l[6]), str_of_tex_ratio(l[0], l[6]))

def count_lines(level, filename):
    filename_src = os.path.join(src(level), filename)
    filename_tgt = os.path.join(tgt(level), filename)
    if os.path.isfile(filename_src):
        wc_src = (int)(check_output("cat %s | wc -l" % filename_src, shell=True)[1])
        (retcode, diff) = check_output("diff %s %s" % (filename_src, filename_tgt), shell=True)

        lines = diff.split("\n")
        removed = len([l for l in lines if l.startswith("< ")])
        added = len([l for l in lines if l.startswith("> ")])
        newed = len([l for l in lines if l.startswith("> (* new *)")])
        return (wc_src, removed, added - newed, newed)
    else:
        wc = int(check_output("cat %s | wc -l" % filename_tgt, shell=True)[1])
        wc_copy = int(check_output("cat %s | grep \"(\\* copy \\*)\" | wc -l" % filename_tgt, shell=True)[1])
        return (0, 0, wc_copy, wc - wc_copy)

def get_result(level, metatheories, passes):
    filenames = subprocess.check_output("cd %s; find . -name '*.v'" % tgt(level), shell=True)
    filenames += subprocess.check_output("cd %s; find . -name '*.vp'" % tgt(level), shell=True)
    metatheories_result = []
    passes_result = []
    for filename in filenames.split("\n"):
        if not filename: continue
        filename = filename[2:]
        lines = count_lines(level, filename)
        if lines:
            if filename in metatheories:
                metatheories_result.append((filename, lines))
            elif filename in passes:
                passes_result.append((filename, lines))
            else:
                raise "strange!"

    metatheories_result.sort(key=lambda x: (x[1][0] == 0, x[1][2] + x[1][3] == 0, x[0], x[1]))
    passes_result.sort(key=lambda x: passes.index(x[0]))
    return (metatheories_result, passes_result)

def print_txt(level, metatheories_result, passes_result):
    with open('results.%s.txt' % level, 'w') as f:
        f.write("Compiler & Metatheories\n")
        others = 0
        for (filename, lines) in metatheories_result:
            if lines[1] == 0 and lines[2] == 0 and lines[3] == 0:
                others += lines[0]
            else:
                f.write("%s\t%s\n" % (filename, str_of_lines(lines)))
        f.write("others\t%d\n" % others)
        f.write("\n")

        f.write("Pass Verifications\n")
        arm = (0, 0, 0, 0)
        ia32 = (0, 0, 0, 0)
        powerpc = (0, 0, 0, 0)
        for (filename, lines) in passes_result:
            if lines[1] == 0 and lines[2] == 0 and lines[3] == 0:
                if filename.startswith("./arm"):
                    arm = add_lines(arm, lines)
                    continue
                elif filename.startswith("./ia32"):
                    ia32 = add_lines(ia32, lines)
                    continue
                elif filename.startswith("./powerpc"):
                    powerpc = add_lines(powerpc, lines)
                    continue

            f.write("%s\t%s\n" % (filename, str_of_lines(lines)))

        f.write("%s\t%s\n" % ("./arm/*proof*.v", str_of_lines(arm)))
        f.write("%s\t%s\n" % ("./ia32/*proof*.v", str_of_lines(ia32)))
        f.write("%s\t%s\n" % ("./powerpc/*proof*.v", str_of_lines(powerpc)))

def lookup_lines(result, filename):
    for (f, l) in result:
        if filename == f:
            return l
    return (0, 0, 0, 0)

def print_tex(metatheories_result1, passes_result1, metatheories_result2, passes_result2):
    passes_acc = {i: (0, 0, 0, 0, 0, 0, 0) for (_, i, _) in PASS_CLASSIFIERS}
    passes_acc[None] = (0, 0, 0, 0, 0, 0, 0)
    metatheories_acc = {i: (0, 0, 0, 0, 0, 0, 0) for (_, i, _) in METATHEORY_CLASSIFIERS}
    metatheories_acc[None] = (0, 0, 0, 0, 0, 0, 0)
    
    for (filename, lines2) in passes_result2:
        lines1 = lookup_lines(passes_result1, filename)
        if lines1[0] != lines2[0]:
            raise "strange!"
        flag = False
        for (pat, i, _) in PASS_CLASSIFIERS:
            if re.match(pat, filename):
                passes_acc[i] = add_lines7(passes_acc[i], (lines1[0], lines1[1], lines1[2], lines1[3], lines2[1], lines2[2], lines2[3]))
                flag = True
                break
        if not flag:
            if not (lines1[1] == 0 and lines2[1] == 0):
                raise "strange!"
            passes_acc[None] = add_lines7(passes_acc[None], (lines1[0], lines1[1], lines1[2], lines1[3], lines2[1], lines2[2], lines2[3]))

    for (filename, lines2) in metatheories_result2:
        lines1 = lookup_lines(metatheories_result1, filename)
        if lines1[0] != lines2[0]:
            raise "strange!"
        flag = False
        for (pat, i, _) in METATHEORY_CLASSIFIERS:
            if re.match(pat, filename):
                metatheories_acc[i] = add_lines7(metatheories_acc[i], (lines1[0], lines1[1], lines1[2], lines1[3], lines2[1], lines2[2], lines2[3]))
                flag = True
                break
        if not flag:
            if not (lines1[1] == 0 and lines2[1] == 0):
                raise "strange!"
            metatheories_acc[None] = add_lines7(metatheories_acc[None], (lines1[0], lines1[1], lines1[2], lines1[3], lines2[1], lines2[2], lines2[3]))

    passes_subtotal = acc_lines7(passes_acc)
    metatheories_subtotal = acc_lines7(metatheories_acc)
    total = add_lines7(passes_subtotal, metatheories_subtotal)

    with open('results.table.tex', 'w') as f:
        f.write("\\begin{tabular}{|l||r||r|r|r||r|r|r|}\n")
        f.write("\\hline\n")
        f.write("& \\multicolumn{1}{c||}{\\textbf{CompCert}} & \\multicolumn{3}{c||}{\\textbf{\\sepcomp-LevelA}} & \\multicolumn{3}{c|}{\\textbf{\\sepcomp-LevelB}} \\\\\n")
        f.write("\\cline{3-8} \n")

        f.write(" & \\multicolumn{1}{c||}{\\textbf{(LOC)}} & \\multicolumn{1}{c|}{\\textbf{Rm}} & \\multicolumn{1}{c|}{\\textbf{AddD}} & \\multicolumn{1}{c||}{\\textbf{AddN}}" + \
                                                         " & \\multicolumn{1}{c|}{\\textbf{Rm}} & \\multicolumn{1}{c|}{\\textbf{AddD}} & \\multicolumn{1}{c|}{\\textbf{AddN}} \\\\\n")
        f.write("\\hline \n")
        f.write("\\textbf{Total} & %s \\\\\n" % (str_of_tex_lines(total, [])))
        f.write("\\hline \n")

        f.write("\\hline\n")
        f.write("\\textbf{Compiler \& Verification} & %s \\\\\n" % (str_of_tex_lines(passes_subtotal, [])))
        f.write("\\hline \n")
        for (_, i, shades) in PASS_CLASSIFIERS:
            f.write("\\texttt{%s} & %s \\\\\n" % (i, str_of_tex_lines(passes_acc[i], shades)))
        f.write("\\specialcell{others in \\texttt{cfrontend}, \\texttt{backend}, \\\\ ~~\\texttt{driver}, \\texttt{arm}, \\texttt{ia32}, \\texttt{powerpc}} & %s \\\\\n" % (str_of_tex_lines(passes_acc[None], [])))
        f.write("\\hline\n")

        f.write("\\hline\n")
        f.write("\\textbf{Metatheory} & %s \\\\\n" % (str_of_tex_lines(metatheories_subtotal, [])))
        f.write("\\hline \n")
        # for (_, i, shades) in METATHEORY_CLASSIFIERS:
        #     f.write("\\texttt{%s} & %s \\\\\n" % (i, str_of_tex_lines(metatheories_acc[i], shades)))
        # f.write("others & %s \\\\\n" % (str_of_tex_lines(metatheories_acc[None], [])))
        # f.write("\\hline\n")

        f.write("\\end{tabular}\n")

if __name__ == "__main__":
    metatheories = subprocess.check_output("cat metatheories.txt", shell=True).split("\n")
    passes = subprocess.check_output("cat passes.txt", shell=True).split("\n")
    (metatheories_result1, passes_result1) = get_result("A", metatheories, passes)
    (metatheories_result2, passes_result2) = get_result("B", metatheories, passes)

    print_tex(metatheories_result1, passes_result1, metatheories_result2, passes_result2)

    # print_txt("A", metatheories_result1, passes_result1)
    # print_txt("B", metatheories_result2, passes_result2)
